package de.lmu.ifi.medien.control;

import java.awt.Component;
import java.awt.Dimension;
import java.util.Vector;

import javax.media.CaptureDeviceInfo;
import javax.media.CaptureDeviceManager;
import javax.media.Codec;
import javax.media.ConfigureCompleteEvent;
import javax.media.ControllerEvent;
import javax.media.ControllerListener;
import javax.media.EndOfMediaEvent;
import javax.media.Format;
import javax.media.Manager;
import javax.media.MediaLocator;
import javax.media.Player;
import javax.media.Processor;
import javax.media.RealizeCompleteEvent;
import javax.media.Time;
import javax.media.control.FormatControl;
import javax.media.control.TrackControl;
import javax.media.format.RGBFormat;
import javax.media.format.VideoFormat;
import javax.media.protocol.CaptureDevice;
import javax.media.protocol.DataSource;
import javax.swing.UIManager;

import com.sun.media.codec.video.cinepak.NativeDecoder;

import de.lmu.ifi.medien.ui.VideoFrame;

/**
 * The main class.
 * @author Sebastian Boring
 */
public class VideoController implements ControllerListener {
	private Processor processor = null;
	private Player player = null;
	
	private LowerThirdCodecControl codecControl;
	private RGBFormat rgbFormat = new RGBFormat(new Dimension(640, 480), 921600, Format.byteArray, 15.0f, 24, 3, 2, 1, 3, 1920, Format.TRUE, RGBFormat.LITTLE_ENDIAN);
	
	private VideoFrame videoFrame;
	
	/**
	 * The constructor
	 */
	public VideoController()
	{
		videoFrame = new VideoFrame(this);
	}
	
	/**
	 * Returns the first capture device that matches the RGB format.
	 * @return A capture device that matches the RGB format, <code>null</code> otherwise.
	 */
	private CaptureDeviceInfo getCaptureDevice()
	{
		Vector devices = CaptureDeviceManager.getDeviceList(rgbFormat);
		if(devices != null && !devices.isEmpty())
		{			
			return (CaptureDeviceInfo)devices.elementAt(0);
		}
		else
		{
			return null;
		}
	}
	
	/**
	 * Tests wether a given format matches one of the supported ones.
	 * @param format The format to be tested.
	 * @param supported The supported formats.
	 * @return A format of the supported formats that matches the given one, <code>null</code> otherwise.
	 */
	private Format matches(Format format, Format[] supported)
	{
        if(supported == null)
        {
            return null;
        }
        for(int i=0; i<supported.length; i++)
        {
            if(supported[i].matches(format))
            {            	
                return supported[i];
            }
        }
        return null;
    }
	
	/**
	 * Initializes a live video stream (e.g. with a webcam).
	 */
	public void initialize()
	{
		CaptureDeviceInfo captureDevice = getCaptureDevice();
		try
		{
			MediaLocator locator = captureDevice.getLocator();
			DataSource dataSource = Manager.createDataSource(locator);
			
			if(!(dataSource instanceof CaptureDevice))
			{
				System.err.println("Could not create DataSource!");
				shutdown(-1);
			}
			
			boolean formatSet = false;
			
			FormatControl[] formatControls = ((CaptureDevice)dataSource).getFormatControls();
			for(int i=0; i<formatControls.length; i++)
			{
	            if(formatControls[i] == null)
	            {
	            	continue;
	            }
	            
	            Format[] formats = formatControls[i].getSupportedFormats();
	            if(formats == null)
	            {
	            	formatSet = false;
	            	break;
	            }

	            if (matches(rgbFormat, formats) != null) {
	            	formatControls[i].setFormat(rgbFormat);
	            	formatSet = true;
	            	
	            	break;
	            }
	        }
			
			if(!formatSet)
			{
				System.err.println("Could not set the correct format!");
				shutdown(-1);
			}
			
			dataSource.connect();
			
			processor = Manager.createProcessor(dataSource);
			processor.addControllerListener(this);
			processor.configure();
		}
		catch(Exception e)
		{
			System.err.println("Could not find any capture devices...");
			shutdown(-1);
		}
	}
	
	/**
	 * Initializes a video stream from a given file.
	 * @param fileName The path and name of the file to be played.
	 */
	public void initialize(String fileName)
	{
		try {
			MediaLocator locator = new MediaLocator("file:" + fileName);
			DataSource dataSource = Manager.createDataSource(locator);
			
			dataSource.connect();
			
			processor = Manager.createProcessor(dataSource);
			processor.addControllerListener(this);
			processor.configure();	
		} catch (Exception e) {
			e.printStackTrace();
		}
	}
	
	/**
	 * Sets up the video frame.
	 * @param visual The player's visual component.
	 * @param control The player's control component.
	 * @param codec The codec's control component.
	 */
	protected void initializeFrame(Component visual, Component control, Component codec)
	{
		videoFrame.addVisualComponent(visual);
		videoFrame.addControlComponent(control);
		videoFrame.addCodecComponent(codec);
		
		visual.setPreferredSize(rgbFormat.getSize());
		visual.setSize(rgbFormat.getSize());
		
		videoFrame.setVisible(true);
		videoFrame.setResizable(false);
		videoFrame.pack();
	}
	
	/**
	 * Implementation of the only method of the ControllerListener interface.
	 * @param evt The ControllerEvent to be handled. 
	 */
	public synchronized void controllerUpdate(ControllerEvent evt)
	{
		if(evt instanceof ConfigureCompleteEvent)
		{
			TrackControl[] tracks = processor.getTrackControls();
			boolean found = false;
			
			for(int i=0; i<tracks.length; i++)
			{
				if(!found && tracks[i].getFormat() instanceof VideoFormat)
				{
					Codec[] videoConversion = new Codec[] {
							new NativeDecoder(), 
							new LowerThirdCodec()};
					codecControl = (LowerThirdCodecControl)(videoConversion[1].getControl("LowerThirdCodecControl"));

					try
					{
						tracks[i].setCodecChain(videoConversion);
						found = true;
					}
					catch(Exception e)
					{
						System.err.println("Could not set the conversion codec!");
					}
				}
			}
			
			if(found)
			{
				processor.realize();
			}
			else
			{
				System.err.println("Could not set the conversion codec!");
			}
		}
		else if(evt instanceof RealizeCompleteEvent)
		{
			try
			{
				player = Manager.createRealizedPlayer(processor.getDataOutput());
				initializeFrame(player.getVisualComponent(), player.getControlPanelComponent(), codecControl.getControlComponent());
				
				player.start();
				processor.start();
			}
			catch(Exception e)
			{
				System.err.println("Could not create a realized player!");
			}
		}
		else if(evt instanceof EndOfMediaEvent)
		{
			player.stop();
			processor.stop();
			
			player.setMediaTime(new Time(0));
			processor.setMediaTime(new Time(0));
			
			player.start();
			processor.start();
		}
	}
	
	/**
	 * Shuts down the application.
	 * @param reason 0 for correct shutdown, -1 otherwise
	 */	
	public void shutdown(int reason)
	{
		try
		{
			processor.stop();
			processor.deallocate();
			
			player.stop();
			player.deallocate();
		}
		catch(Exception e) { }
		finally
		{
			System.exit(reason);
		}
	}
	
	public static void main(String[] args)
	{
		try
		{
			if(args.length != 1)
			{
				System.err.println("Usage: java VideoController [fileName]");
				System.err.println("[fileName]:\tPath and name of the video file...");
				
				System.exit(0);
			}
			else
			{			
				UIManager.setLookAndFeel(UIManager.getSystemLookAndFeelClassName());
			
				VideoController controller = new VideoController();
				controller.initialize(args[0]);
			}
		}
		catch(Exception e)
		{
			System.exit(-1);
		}
	}
}
